---> RTK_BANDS (body) <--- %W% %G%  -- MVME133a Version.

-- Copyright (C) 1991, the Florida State University.

-- This code was produced by FSU students and faculty, working with
-- support from the Florida High Technology Council.

-- Permission to copy and distribute this code entire and to execute
-- it within a computer, without fee, is granted provided that this
-- is not done for commercial advantage, and that this notice of FSU
-- copyright is included.  Any other form of use requires negotiated
-- permission of the Florida State University.

-- Original Authors:

-- T.P. Baker
-- Chi-Sing Chen
-- Pratit Santiprabhob

-- Modifications: (in chronological order, oldest first)

-- 1. Prescaler

-- Date 27 September 1991
-- Author: Pratit Santiprabhob

-- A Prescaler is added, so that we can vary the speed of the timer
-- in seven levels instead of the previous two.

with UTIL,
     KEYBOARD_INPUT,
     SCREEN_OUTPUT,
     SYSTEM,
     ANSI_SYS,
     TIMER,
     Simplest_IO,
     UNCHECKED_CONVERSION;
  pragma ELABORATE(RTK);
  pragma ELABORATE(KEYBOARD_INPUT);
  pragma ELABORATE(SCREEN_OUTPUT);
  pragma ELABORATE(TIMER);
package body RTK_BANDS is
  use UTIL;
  use RTK;
  use KEYBOARD_INPUT;
  use SCREEN_OUTPUT;
  use ANSI_SYS;
  use TIMER;
  use SYSTEM;
  use Simplest_IO;
  procedure HALT renames UTIL.HALT;
  function Add_to_Band_Range is new UNCHECKED_CONVERSION(System.Address,BAND_RANGE);
  function Band_Range_to_Add is new UNCHECKED_CONVERSION(BAND_RANGE,System.Address);

  type Stack is array(1..1024) of Integer;
  Stack_1, Stack_2, Stack_3, Stack_4, Stack_5, Stack_6, Stack_7: Stack;
  Integer_Size: constant:= Integer'size/System.Storage_Unit;
  Init_1: Init_State:= (Stack_1'address, Stack_1'length*Integer_Size);
  Init_2: Init_State:= (Stack_2'address, Stack_2'length*Integer_Size);
  Init_3: Init_State:= (Stack_3'address, Stack_3'length*Integer_Size);
  Init_4: Init_State:= (Stack_4'address, Stack_4'length*Integer_Size);
  Init_5: Init_State:= (Stack_5'address, Stack_5'length*Integer_Size);
  Init_6: Init_State:= (Stack_6'address, Stack_6'length*Integer_Size);
  Init_7: Init_State:= (Stack_7'address, Stack_7'length*Integer_Size);

  type BAND_BIT_VECTOR is array (BAND_RANGE) of BOOLEAN;
  Cycle: constant:= 16380;
  BEFORE_NOTHING: constant:= -CYCLE;
  BEFORE_EVERYTHING: constant:= 2*CYCLE;
  SOUND_ON: BOOLEAN:= TRUE;
  TICK_COUNT: Natural:= 0;
  ENABLED: BAND_BIT_VECTOR:=(others=> TRUE);
  MUTEX: BAND_BIT_VECTOR:= (others=> FALSE); -- use R(I), or not
  LOAD: array (BAND_RANGE) of INTEGER:=(1,2,3,4,5);
  DEADLINE:  array (BAND_RANGE) of INTEGER:=(others=> 1);
  DONE: array (BAND_RANGE) of INTEGER:=(others=> BEFORE_EVERYTHING);
  R: ARRAY (BAND_RANGE) OF RTK.LOCK_ID; -- lock for "dining philosphers"
  P: INTEGER range 0..4 :=0;            -- cursor band position
  T: BAND_RANGE:=BAND_RANGE'first;      -- cursor band number
  COL_OFFSET: constant:=18;             -- scanning column
  ROW: LINE_NUMBER:=1;                  -- cursor row
  CHAR: array (LINE_NUMBER) of CHARACTER:= (others=> ' ');
  FIRST_BAND_PRIORITY: constant:= 0;
  LAST_MARKED_TICK: INTEGER:= 0;
  MAX_DIFF: INTEGER:= 0;
  MAX_NEST: INTEGER:= 0;
  All_Done: Boolean:= False;
  Tick_Dummy: Natural:= 0;
  Prescaler: Timer.Prescaler_Value:= Timer.Two_Hundred;
    -- Prescaler for Timer D.

  Init_Screen_ID: RTK.VP_ID;

  Started: array(BAND_RANGE) of Boolean:= (others => False);

  Main_VP_ID: VP_ID;

  function IS_BEFORE(T1,T2: INTEGER) return BOOLEAN is
     -- true iff T1 is before T2 on CLOCK, assuming T1 and T2
     -- are not separated by more than 1/2 cycle
     D: INTEGER:= T1-T2;
     HALF: constant:= CYCLE/2;
    --
    --  false  ]  true   [ false  ]  true
    -----------|---------|---------|-----------
    --       -1/2        0        1/2
    begin
      return D > HALF or else (D <= 0 and then D > -HALF);
    end IS_BEFORE;

  function UP_ARROW return CHARACTER is  -- same as tarte's
  begin return ASCII.DC1;
  end UP_ARROW;

  function DOWN_ARROW return CHARACTER is   -- same as tarte's
  begin return ASCII.DC2;
  end DOWN_ARROW;

  procedure FILL_COLUMN(CH: CHARACTER) is  -- ok
    ROW: LINE_NUMBER:=1;
  begin
    loop
      PUT_CHAR(CH,ROW,COLUMN_NUMBER(((TICK_COUNT+1) mod 60)+COL_OFFSET));
      exit when ROW= Line_Number'last;
      ROW:=ROW+1;
    end loop;
  end FILL_COLUMN;

  function ONE_CHR(INT: INTEGER) return CHARACTER is   --ok
  begin if INT<=9 then return CHARACTER'val(CHARACTER'pos('0')+INT);
        else return CHARACTER'val(CHARACTER'pos('A')+INT-10); end if;
  end ONE_CHR;

  procedure TICKER is
    use RTK;
    C,D: INTEGER;
  begin
     if Tick_Dummy /= 5
     then Tick_Dummy:= Tick_Dummy + 1;
          return;
     else Tick_Dummy:= 0;
     end if;
      TICK_COUNT:= (TICK_COUNT+1) mod CYCLE;
      for I in BAND_RANGE loop
        if TICK_COUNT = DEADLINE(I) then
          if IS_BEFORE(TICK_COUNT,DONE(I)) then -- missed deadline
            PUT(ASCII.BEL);
            PUT_MISSED_DEADLINE_MARK
              (LINE_NUMBER(I)*5-4,
               COLUMN_NUMBER(((DEADLINE(I)-1) mod 60)+COL_OFFSET));
          else
            PUT_DEADLINE_MARK
              (LINE_NUMBER(I)*5-4,
               COLUMN_NUMBER(((DEADLINE(I)-1) mod 60)+COL_OFFSET));
          end if;
          DONE(I):= BEFORE_NOTHING;
          -- schedule next request, if necessary
          if ENABLED(I) then 
            DEADLINE(I):= (DEADLINE(I)+PERIOD(I)) mod CYCLE;
            RTK.RESUME_VP(ID(I), 1);
          end if;
        end if; -- at deadline   
      end loop; -- for I in BAND_RANGE
      FILL_COLUMN(' '); 
    RTK.RESUME_VP(KBD_ID, 1);
  exception
  when others=>
     PUT_LINE("Failed in TICKER !!");
     HALT;
  end TICKER;

  function CURRENT_RTK_ATT return BAND_RANGE is
    use RTK;
  begin
    return Add_to_Band_Range(RTK.Get_Attribute(RTK.SELF,1));
  end CURRENT_RTK_ATT;

  procedure VP_Int_Handler is
  use RTK;
  begin
    RTK.Suspend_Self(1);
  end VP_Int_Handler;

  procedure BAND is
    use RTK;
    T: BAND_RANGE;
    ROW_OFFSET: INTEGER;
    I: INTEGER;
    LOCKS_SET: BOOLEAN:= FALSE;
  begin
    T:= CURRENT_RTK_ATT;
    if T=0 then HALT; end if;
    if not Started(T) then
    RTK.Attach_Interrupt_Handler(Interrupt=> RTK.VP_Interrupt_ID'first,
                                 Priority=> Interrupt_Priority'last - 9,
                                 Handler_Address=> VP_Int_Handler'Address,
                                 Info=> 0);
    Started(T):= True;
    end if;
    ROW_OFFSET:= (INTEGER(T)-1)*5+1; I:=1;
    loop -- this loop is (only) to model increasing load.
      if MUTEX(T) then
        if T=5 then RTK.SEIZE_LOCK(R(1)); RTK.SEIZE_LOCK(R(5));
        else RTK.SEIZE_LOCK(R(T+1)); RTK.SEIZE_LOCK(R(T));
        end if;
        LOCKS_SET:= TRUE;
      end if;
      for L in 1..INTEGER(2**(I-1)) loop
        PUT_STAGE_MARK((I-1)/4,
          LINE_NUMBER((I-1) mod 4+1)+ROW_OFFSET,
          COLUMN_NUMBER((TICK_COUNT mod 60)+COL_OFFSET));
      end loop;
      if LOCKS_SET then
        if T=5 then RTK.RELEASE_LOCK(R(5)); RTK.RELEASE_LOCK(R(1));
        else RTK.RELEASE_LOCK(R(T)); RTK.RELEASE_LOCK(R(T+1));
	        end if;
        LOCKS_SET:= FALSE;
      end if;
      exit when I>=LOAD(T); 
      I:=I+1;
    end loop;
    DONE(T):= TICK_COUNT;
  exception
   when others=>
    PUT_LINE("Failed in BAND !! ");
    HALT;
    DONE(T):= TICK_COUNT;
  end BAND;

  procedure INIT_SCREEN is
  begin
    CLEAR_SCREEN;
    -- Put left margin labels on screen.
    ROW:=1;
    for I in BAND_RANGE loop
      ROW:=ROW+1;
      PUT_STRING("PRIORITY=",ROW,1);
      CHAR(ROW):=ONE_CHR(INTEGER(RTK.BASE_PRIORITY(RTK_BANDS.ID(I)))
        -FIRST_BAND_PRIORITY);
      PUT_CHAR(CHAR(ROW),ROW,16);
      ROW:=ROW+1;
      PUT_STRING("PERIOD=  ",ROW,1);
      CHAR(ROW):=ONE_CHR(INTEGER(PERIOD(I)));
      PUT_CHAR(CHAR(ROW),ROW,16);
      ROW:=ROW+1;
      PUT_STRING("LOAD=    ",ROW,1);
      CHAR(ROW):=ONE_CHR(LOAD(I));
      PUT_CHAR(CHAR(ROW),ROW,16);
      ROW:=ROW+1;
      CHAR(ROW):='R';
      PUT_CHAR('R',ROW,16);
      if ROW<LINE_NUMBER'last then ROW:=ROW+1; end if;
    end loop;
    ROW:=1;
    PUT_REV_CHAR(CHAR(ROW),ROW,16);
  end INIT_SCREEN;

  procedure KBD_SERVICE is
    CH: CHARACTER;
    I: INTEGER;
  begin
     if not KEYBOARD_INPUT.KBD_DATA_AVAILABLE then return; end if;
     CH:= KEYBOARD_INPUT.READ_KBD_NO_ECHO;
     if CH/=ASCII.NUL then
        case CH is
        when '1'..'9'|'a'..'c'|'A'..'C'=>
        -- change thread at cursor position to new priority --
        begin 
           if CH in 'a'..'c' then
              CH:=CHARACTER'val(CHARACTER'pos('A')+
                  CHARACTER'pos(CH)-CHARACTER'pos('a'));
           end if;
           if CH in 'A'..'C'then
              I:=CHARACTER'pos(CH)-CHARACTER'pos('A')+10;
           else I:=CHARACTER'pos(CH)-CHARACTER'pos('0');
           end if;
           case P is
           when 1=>
              CHAR(ROW):=CH;
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
              RTK.SET_BASE_PRIORITY
                (RTK_BANDS.ID(T),
                 VP_Priority(I+FIRST_BAND_PRIORITY));
           when 2=>
              CHAR(ROW):=CH;
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
              PERIOD(T):=I;
           when 3=>
              CHAR(ROW):=CH;
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
              LOAD(T):=I;
           when others=>
              null;
           end case;
        end;
        when '^'=> -- raise clock rate
              if Prescaler /= Timer.Prescaler_Value'first
              then Prescaler:= Timer.Prescaler_Value'pred(Prescaler);
                   STOP_TIMER(Timer.Timer_D); 
                   START_TIMER(Timer.Timer_D, Prescaler,
                               Timer.Twenty_Msec_Mode);
              end if;
        when 'v'=> -- lower clock rate
              if Prescaler /= Timer.Prescaler_Value'last
              then Prescaler:= Timer.Prescaler_Value'succ(Prescaler);
                   STOP_TIMER(Timer.Timer_D); 
                   START_TIMER(Timer.Timer_D, Prescaler,
                               Timer.Twenty_Msec_Mode);
              end if;
        when 's' | 'S'=> -- suspend thread at cursor position --
           if P=4 then
              CHAR(ROW):='S';
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
              ENABLED(T):= FALSE; DEADLINE(T):= BEFORE_NOTHING;
	      RTK.INTERRUPT_VP(ID(T),RTK.VP_Interrupt_ID'first,0);
           end if;
        when 'r' | 'R'=>  -- release thread at cursor position --
           if P=4 then
              CHAR(ROW):='R';
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
              if not ENABLED(T) then
                ENABLED(T):= TRUE;
                DEADLINE(T):= (TICK_COUNT + PERIOD(T)) mod CYCLE;
                DONE(T):= BEFORE_NOTHING;
                RTK.RESUME_VP(RTK_Bands.ID(T),1);
              end if;
           end if;
        when 't' | 'T'=>  -- toggle sound switch --
           SOUND_ON:= not SOUND_ON;
        when 'm' | 'M' | 'd' | 'D' => -- toggle dining --
           if P=4 then
              MUTEX(T):= not MUTEX(T);
              if MUTEX(T) then PUT_STRING("[DINING]",ROW,1);
              else PUT_STRING("        ",ROW,1); end if;
              end if;
        when 'q' | 'Q' | ASCII.ETX =>
           KEYBOARD_INPUT.CLOSE_KEYBOARD;
           All_Done:= True;
           for J in Band_Range loop
             RTK.Destroy_VP(RTK_Bands.ID(J));
           end loop;
           RTK.Destroy_VP(Init_Screen_ID);
           RTK.Detach_Interrupt_Handler(Interrupt=> Timer.Timer_D_Vector);
           -- just for testing.
           RTK.Destroy_VP(KBD_ID);
        when others=>
           if CH=UP_ARROW then -- move cursor up --
              PUT_CHAR(CHAR(ROW),ROW,16);
              if P=0 then
                 P:=4;
                 if T=BAND_RANGE'first then
                    T:=BAND_RANGE'last; ROW:= Line_Number'last;
                 else T:=T-1; ROW:=ROW-1;
                 end if;
              else P:=P-1; ROW:=ROW-1;
              end if;
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
           elsif CH=DOWN_ARROW then -- move cursor down --
              PUT_CHAR(CHAR(ROW),ROW,16);
              if P=4 then
                 P:=0;
                 if T=BAND_RANGE'last then
                    T:=BAND_RANGE'first; ROW:=1;
                 else T:= T+1; ROW:=ROW+1;
                 end if;
              else P:=P+1; ROW:=ROW+1;
              end if;
              PUT_REV_CHAR(CHAR(ROW),ROW,16);
           end if;
        end case;
     end if;
  exception
  when others=>
     PUT_LINE("Failed in KEYBOARD !!");
     HALT;
  end KBD_SERVICE;

  procedure HANDLE_FAILURE is
  begin
    HALT;
  end HANDLE_FAILURE;

  procedure IDLE is
  begin null;
  end IDLE;

begin
  RTK.INITIALIZE_LOCK(R(1), FIRST_BAND_PRIORITY+16);
  RTK.INITIALIZE_LOCK(R(2), FIRST_BAND_PRIORITY+16);
  RTK.INITIALIZE_LOCK(R(3), FIRST_BAND_PRIORITY+16);
  RTK.INITIALIZE_LOCK(R(4), FIRST_BAND_PRIORITY+16);
  RTK.INITIALIZE_LOCK(R(5), FIRST_BAND_PRIORITY+16);
  RTK.CREATE_VP
     (PRIORITY=> FIRST_BAND_PRIORITY+5,
      INITIAL_STATE=>Init_1,
      ENTRY_POINT=> RTK_Bands.Band'address,
      VP=> RTK_Bands.ID(1));
  RTK.CREATE_VP
     (PRIORITY=> FIRST_BAND_PRIORITY+4,
      INITIAL_STATE=>Init_2,
      ENTRY_POINT=> RTK_Bands.Band'address,
      VP=> RTK_Bands.ID(2));
  RTK.CREATE_VP
     (PRIORITY=> FIRST_BAND_PRIORITY+3,
      INITIAL_STATE=>Init_3,
      ENTRY_POINT=> RTK_Bands.Band'address,
      VP=> RTK_Bands.ID(3));
  RTK.CREATE_VP
     (PRIORITY=> FIRST_BAND_PRIORITY+2,
      INITIAL_STATE=>Init_4,
      ENTRY_POINT=> RTK_Bands.Band'address,
      VP=> RTK_Bands.ID(4));
  RTK.CREATE_VP
     (PRIORITY=> FIRST_BAND_PRIORITY+1,
      INITIAL_STATE=>Init_5,
      ENTRY_POINT=> RTK_Bands.Band'address,
      VP=> RTK_Bands.ID(5));
  RTK.CREATE_VP
     (PRIORITY=> VP_Priority'last,
      INITIAL_STATE=>Init_6,
      ENTRY_POINT=> RTK_Bands.KBD_Service'address,
      VP=> RTK_Bands.KBD_ID);
  RTK.Create_VP
     (Priority=> VP_Priority'last,
      Initial_State=> Init_7,
      Entry_Point=> RTK_Bands.Init_Screen'address,
      VP=> RTK_Bands.Init_Screen_ID);
    RTK.Set_Attribute(RTK_Bands.ID(1),1,Band_Range_to_Add(1));
    RTK.Set_Attribute(RTK_Bands.ID(2),1,Band_Range_to_Add(2));
    RTK.Set_Attribute(RTK_Bands.ID(3),1,Band_Range_to_Add(3));
    RTK.Set_Attribute(RTK_Bands.ID(4),1,Band_Range_to_Add(4));
    RTK.Set_Attribute(RTK_Bands.ID(5),1,Band_Range_to_Add(5));
  RTK_BANDS.PERIOD:= (2,3,5,10,12);
  RTK.Attach_Interrupt_Handler(Interrupt=> Timer.Timer_D_Vector,
	                       Priority=> Interrupt_Priority'last,
	                       Handler_Address=> Ticker'address,
	                       Info=> 0);
  RTK.Resume_VP(RTK_Bands.Init_Screen_ID, 1);
  START_TIMER(Timer.Timer_D, Prescaler, Timer.Twenty_Msec_Mode);
  Main_VP_ID:= RTK.Self;
  RTK.Destroy_VP(Main_VP_ID);
  -- just for testing.
--  while not All_Done loop IDLE; end loop;
--  RTK.Detach_Interrupt_Handler(Interrupt=> Timer.Timer_D_Vector);
--  No need, since the Main_VP which is this procedured is
--  destroyed, any way.
end RTK_BANDS;
